#!/usr/bin/python3
# -*- coding: utf-8 -*-
# 詰将棋解答プログラム
# jshogi.py Ver.1.00
#  本田 博通 2024/10/27
#  使用条件: GPL v.3
#  python 3.8 上で動作チェック
#  mshogi改良。numpy は使わない。
#  打歩詰めの禁に対応。9手詰めを 3秒程度で解ける。
#  日本将棋連盟のまいにち詰将棋 11-13手詰めも短時間でほぼ解ける。
#  合駒はある毎に、+2手の応酬を数えるので、実際の詰手数とずれることがある。
import re
import copy

NAME = []
stat = []
rname = ["","１","２","３","４","５","６","７","８","９"]
cname = ["","一","二","三","四","五","六","七","八","九"]
mochic2name0 = ['hisha', 'kaku', 'kin', 'gin', 'kei', 'kyo', 'fu']
kname = {'o':"王",'hisha':"飛",'kaku':"角",'kin':"金",'gin':"銀",'kei':"桂",
         'kyo':"香",'fu':"歩",'r_o':"王",'r_hisha':"龍",'r_kaku':"馬",'r_kin':"金",
         'r_gin':"成銀",'r_kei':"成桂",'r_kyo':"成香",'r_fu':"と"}
able = {'fu':[[[0,-1]]], # 単純な1手順
        'r_fu':[[[1,-1]],[[0,-1]],[[-1,-1]],[[1,0]],[[-1,0]],[[0,1]]], # 独立した6手
        'kyo':[[[0,-1],[0,-2],[0,-3],[0,-4],[0,-5],[0,-6],[0,-7],[0,-8]]], # 1方向に連動した8手がある
        'r_kyo':[[[1,-1]],[[0,-1]],[[-1,-1]],[[1,0]],[[-1,0]],[[0,1]]],
        'kei':[[[1,-2]],[[-1,-2]]], # 独立した2手
        'r_kei':[[[1,-1]],[[0,-1]],[[-1,-1]],[[1,0]],[[-1,0]],[[0,1]]],
        'gin':[[[1,-1]],[[0,-1]],[[-1,-1]],[[-1,1]],[[1,1]]],
        'r_gin':[[[1,-1]],[[0,-1]],[[-1,-1]],[[1,0]],[[-1,0]],[[0,1]]],
        'kin':[[[1,-1]],[[0,-1]],[[-1,-1]],[[1,0]],[[-1,0]],[[0,1]]],
        'r_kin':[[[1,-1]],[[0,-1]],[[-1,-1]],[[1,0]],[[-1,0]],[[0,1]]],
        'o':[[[1,-1]],[[0,-1]],[[-1,-1]],[[1,0]],[[-1,0]],[[1,1]],[[0,1]],[[-1,1]]],
        'hisha':[[[0,-1],[0,-2],[0,-3],[0,-4],[0,-5],[0,-6],[0,-7],[0,-8]],
                 [[0,1],[0,2],[0,3],[0,4],[0,5],[0,6],[0,7],[0,8]],
                 [[1,0],[2,0],[3,0],[4,0],[5,0],[6,0],[7,0],[8,0]],
                 [[-1,0],[-2,0],[-3,0],[-4,0],[-5,0],[-6,0],[-7,0],[-8,0]]], # 4方向に連動した8手がある
        'r_hisha':[[[1,1]],[[1,-1]],[[-1,1]],[[-1,-1]],
                   [[0,-1],[0,-2],[0,-3],[0,-4],[0,-5],[0,-6],[0,-7],[0,-8]],
                   [[0,1],[0,2],[0,3],[0,4],[0,5],[0,6],[0,7],[0,8]],
                   [[1,0],[2,0],[3,0],[4,0],[5,0],[6,0],[7,0],[8,0]], # 4方向に独立した4手
                   [[-1,0],[-2,0],[-3,0],[-4,0],[-5,0],[-6,0],[-7,0],[-8,0]]], # 4方向に連動した8手
        'kaku':[[[-1,-1],[-2,-2],[-3,-3],[-4,-4],[-5,-5],[-6,-6],[-7,-7],[-8,-8]],
                [[-1,1],[-2,2],[-3,3],[-4,4],[-5,5],[-6,6],[-7,7],[-8,8]],
                [[1,1],[2,2],[3,3],[4,4],[5,5],[6,6],[7,7],[8,8]],
                [[1,-1],[2,-2],[3,-3],[4,-4],[5,-5],[6,-6],[7,-7],[8,-8]]],
        'r_kaku':[[[1,0]],[[-1,0]],[[0,1]],[[0,-1]],
                  [[-1,-1],[-2,-2],[-3,-3],[-4,-4],[-5,-5],[-6,-6],[-7,-7],[-8,-8]],
                  [[-1,1],[-2,2],[-3,3],[-4,4],[-5,5],[-6,6],[-7,7],[-8,8]],
                  [[1,1],[2,2],[3,3],[4,4],[5,5],[6,6],[7,7],[8,8]],
                  [[1,-1],[2,-2],[3,-3],[4,-4],[5,-5],[6,-6],[7,-7],[8,-8]]]}
IO = -1
TEJUN = ""
MAP0 = [[-1,-1,-1,-1,-1,-1,-1,-1,-1,-1],
        [-1,-1,-1,-1,-1,-1,-1,-1,-1,-1],
        [-1,-1,-1,-1,-1,-1,-1,-1,-1,-1],
        [-1,-1,-1,-1,-1,-1,-1,-1,-1,-1],
        [-1,-1,-1,-1,-1,-1,-1,-1,-1,-1],
        [-1,-1,-1,-1,-1,-1,-1,-1,-1,-1],
        [-1,-1,-1,-1,-1,-1,-1,-1,-1,-1],
        [-1,-1,-1,-1,-1,-1,-1,-1,-1,-1],
        [-1,-1,-1,-1,-1,-1,-1,-1,-1,-1],
        [-1,-1,-1,-1,-1,-1,-1,-1,-1,-1]] # 10x10
MAXLEVEL = 7 # +2手詰め
stat0 = []
stat_bk = []

mochi0 = {'hisha':-1, 'kaku':-1, 'kin':-1, 'gin':-1, 'kei':-1, 'kyo':-1, 'fu':-1}

### ここまで最適化完了

def isNari(i, c):
    global stat, NAME
    if stat[i][3] == 1: return True # もとから成
    if stat[i][0] > 10 or re.match('^(o|kin)$', NAME[i]):
        return False # 成がない、盤面外のものは False
    if stat[i][2] == 0 and stat[i][3] == 0: # 先手、不成で
        if stat[i][1] < 4: return True # 相手陣から動く
        elif c < 4: return True # 相手陣へ動く
    elif stat[i][2] == 1 and stat[i][3] == 0: # 後手、不成で
        if stat[i][1] > 6: return True # 相手陣から動く
        elif c > 6: return True # 相手陣へ動く
    return False

def komaSet(query): # 攻撃の王は除く
    global stat, NAME, MAP, aMAP, dMAP, aids, dids, IO
    query = re.sub(r'^.+\?', '', query)
    query = query.replace('\&amp;', '&')
    query = query.rstrip() # 末尾の改行と & を除く
    query = query.rstrip('&')
    if 'fu=' in query: # get で駒配置データがきた場合
        ii = 0
        keys = query.split('&')
        aids = []
        dids = []
        for key in keys:
            if re.match('^[a-z]+=\d+,[0-9-]+,\d,\d$', key):
                val = key.split('=')
                s = val[1].split(',')
                if val[0] == 'o':
                    if s[2] == '0': continue # 攻撃側の王は除く
                NAME.append(val[0])
                [r, c, s, n] = [int(s[0]), int(s[1]), int(s[2]), int(s[3])]
                if c < 0: c = 0
                stat.append([r, c, s, n]) # s=0: 攻撃, 1: 守備
                if val[0] == 'o':
                    IO = ii
                if s == 0:
                    aids.append(ii)
                else:
                    dids.append(ii)
                if r < 10 and c > 0:
                    MAP[r][c] = ii
                    if s == 0:
                        aMAP[r][c] = ii
                    else:
                        dMAP[r][c] = ii
                ii += 1

def isNifu(i, r): # 打つ前に実行
    global aMAP, stat
    if (NAME[i] == "fu"):
        if stat[i][2] == 0: # 先手
            for c in [1,2,3,4,5,6,7,8,9]:
                if (aMAP[r][c] > -1):
                    j = aMAP[r][c]
                    if i != j and NAME[j] == "fu" and stat[j][3] == 0: return True
        else: # 後手は他の駒を打てばいいので、歩打ちは考えなくてよい。
            return False
    return False

def getRname(i):
    global stat, NAME
    return NAME[i] if stat[i][3] == 0 else "r_" + NAME[i]

def getAigomaRC(r0, c0, r, c): # 効率のため常に r0, c0 に近い方からの配列を返す(3まで)
    i = 0
    if r0 == r: # 飛車、香の場合
        if c0 < c:
            if c - c0 > 3: c = c0 + 4
            return [[r, c2]  for c2 in range(c0 + 1, c)]
        else:
            if c0 - c > 3: c = c0 - 4
            return [[r, c2] for c2 in range(c0 - 1, c, -1)]
    elif c0 == c:
        if r0 < r:
            if r - r0 > 3: r = r0 + 4
            return [[r2, c]  for r2 in range(r0 + 1, r)]
        else:
            if r0 - r > 3: r = r0 - 4
            return [[r2, c]  for r2 in range(r0 - 1, r, -1)]
    else: # 角の場合
        if c0 < c and r0 < r:
            if c - c0 > 3:
                c = c0 + 4
                r = r0 + 4
            return [[r2, c2] for r2, c2 in zip(range(r0 + 1, r), range(c0 + 1, c))]
        elif c0 > c and r0 < r:
            if r - r0 > 3:
                c = c0 - 4
                r = r0 + 4
            return [[r2, c2]
                    for r2, c2 in zip(range(r0 + 1, r), range(c0 - 1, c, -1))]
        elif c0 < c and r0 > r:
            if c - c0 > 3:
                c = c0 + 4
                r = r0 - 4
            return [[r2, c2]
                    for r2, c2 in zip(range(r0 - 1, r, -1), range(c0 + 1, c))]
        else:
            if c0 - c > 3:
                c = c0 - 4
                r = r0 - 4
            return [[r2, c2]
                    for r2, c2 in zip(range(r0 - 1, r, -1), range(c0 - 1, c, -1))]
    return [] # ここはない...

def getaRCState(r, c): # その場所へ攻撃駒が進めるか
    global aMAP, dMAP
    # 移動不能は呼び出し元で既にチェック済
    if aMAP[r][c] > -1: return 1 # 味方
    elif dMAP[r][c] > -1: return 0 # 敵
    else: return -1 # 何もない

def getdRCState(r, c): # その場所へ守備駒が進めるか
    global aMAP, dMAP
    # 移動不能は呼び出し元で既にチェック済
    if dMAP[r][c] > -1: return 1 # 味方
    elif aMAP[r][c] > -1: return 0 # 敵
    else: return -1 # 何もない

# 先手番が持駒を持つか。持っていればその種の i の一覧を返す
def getaMochi():
    global aids, stat, NAME
    mochi = copy.copy(mochi0)
    ret = []
    for i in aids:
        if stat[i][0] > 9: mochi[NAME[i]] = i
    return [v for v in mochi.values() if v > -1]

# 後手番が持駒を持つか。持っていればその種の i の一覧を返す
def getdMochi():
    global dids, stat, NAME
    mochi = copy.copy(mochi0)
    ret = []
    for i in dids:
        if NAME[i] == 'fu': continue # 歩は弱いので指す価値がほとんどない
        # あるとしたら、持ち駒がきわめて少ない場合だが詰め将棋にそれはない
        if stat[i][0] > 9: mochi[NAME[i]] = i
    return [v for v in mochi.values() if v > -1]

def getAkiArea(): # あとで効率化!!!!! kei がX
    global IO, MAP
    ret = []
    [ro, co] = [stat[IO][0], stat[IO][1]]
    for rr, cc in zip([0,1,0,-1,1,1,-1,-1,0,2,0,-2,0,3,0,-3,2,2,-2,-2,3,3,-3,-3,2,2,-2,-2,1,1,-1,-1,1,1,-1,-1,3,3,-3,-3,3,3,-3,-3,2,2,-2,-2],[1,0,-1,0,1,-1,1,-1,2,0,-2,0,3,0,-3,0,2,-2,2,-2,3,-3,3,-3,1,-1,1,-1,2,-2,2,-2,3,-3,3,-3,1,-1,1,-1,2,-2,2,-2,3,-3,3,-3]): # 飛が先
        r, c = ro + rr, co + cc
        if r > 0 and r < 10 and c > 0 and c < 10: # 王周辺
            if MAP[r][c] == -1:
                ret.append([r, c])
    return ret

def pushMap(i, r, c, s, n): # コマ i が r, c, s, n の状態となる
    global stat, stat_bk, MAP, aMAP, dMAP, aids, dids
    [r0, c0, s0, n0] = stat[i] # i の以前の値
    stat_bk.append([i, copy.copy(stat[i]), copy.copy(aids), copy.copy(dids)])
    if r0 < 10:
        MAP[r0][c0] = -1
        if s0 == 0:
            aMAP[r0][c0] = -1
        else:
            dMAP[r0][c0] = -1
    if r < 10:
        MAP[r][c] = i
        if s == 0:
            aMAP[r][c] = i
        else:
            dMAP[r][c] = i
    elif s0 != s:
        if s0 == 1: # dids を減らし、aids を増やす
            dids.remove(i)
            aids.append(i)
        else: # aids を減らし、dids を増やす
            aids.remove(i)
            dids.append(i)
    stat[i] = [r, c, s, n]

def popMap(i): # stat, MAP の修正を行う
    global stat, stat_bk, MAP, aMAP, dMAP, aids, dids
    [r, c, s, n] = stat[i] # i の今の値
    [i, stat[i], aids, dids] = stat_bk.pop()
    [r0, c0, s0, n0] = stat[i] # i の変更後の値
    if r < 10:
        MAP[r][c] = -1
        if s == 0:
            aMAP[r][c] = -1
        else:
            dMAP[r][c] = -1
    if r0 < 10:
        MAP[r0][c0] = i
        if s0 == 0:
            aMAP[r0][c0] = i
        else:
            dMAP[r0][c0] = i

def findKomaAuto(r, c): # r, c の位置にいる駒を探す
    global MAP
    if r > 9: return -1
    return MAP[r][c]

def canoGo(r, c): # 王が行けるか
    global IO
    absr, absc = abs(r - stat[IO][0]), abs(c - stat[IO][1])
    if absr < 2 and absc < 2 : return True
    return False

# これは面倒なので、mshogi をそのまま使う
# state:0(敵)、-1(何もいない)、1(Default)、2:味方を敵扱い
# i が行ける場所を [99,98..] といった配列で返す
def getMovableArea(name, r, c, sen, state):
    global able
    cango = []
    if r > 9: return [] # 盤上にないものはとりあえず空配列を返す
    for rcs in able.get(name):
        stop = False
        for [rr, cc] in rcs:
            if stop: break
            if sen > 0: [cc, rr] = [-cc, -rr]
            [r1, c1] = [r + rr, c + cc]
            if c1 < 1 or c1 > 9 or r1 < 1 or r1 > 9: break
            if sen == 0:
                state1 = getaRCState(r1, c1)
            else:
                state1 = getdRCState(r1, c1)
            if len(rcs) == 1: # 1コマしか進まない場合
                if state == 0: # 敵
                    if state1 == 0: cango.append([r1, c1])
                elif state == -1: # 何もいない
                    if state1 == -1: cango.append([r1, c1])
                elif state == 2: # 1 に加え、味方もOK
                    if state1 < 2: cango.append([r1, c1])
                else: # デフォルト
                    if state1 < 1: cango.append([r1, c1])
            else: # どこまでも進む場合
                if state == 0: # 敵
                    if state1 == 0:
                        cango.append([r1, c1])
                        stop = True
                        break
                    elif state1 > 0:
                        stop = True
                        break
                elif state == -1: # 何もいない
                    if state1 > -1:
                        stop = True
                        break
                    else: cango.append([r1, c1])
                elif state == 2: # 1 に加え、味方もOK
                    if state1 == 0 or state1 == 1:
                        cango.append([r1, c1])
                        stop = True
                        break
                    else: cango.append([r1, c1])
                else: # state:1
                    if state1 == 0:
                        cango.append([r1, c1])
                        stop = True
                        break
                    elif state1 == 1 or state1 == 2:
                        stop = True
                        break
                    else: cango.append([r1, c1]) # -1 何もない
    return cango

# mshogi を aids,dids 対応以外そのまま持ってくる。
# sen の全駒が状態 state で行ける範囲を配列で返す
# state:0(敵)、-1(何もいない)、1(Default)、2:味方を敵扱い
def getMovableAreaAll(sen, state):
    global stat, aids, dids
    ilist = []
    if sen == 0:
        for i in aids:
            if stat[i][0] < 10:
                area = getMovableArea(getRname(i), stat[i][0], stat[i][1], 0, state)
                if len(area) != 0: ilist.extend(area)
    else:
        for i in dids:
            if stat[i][0] < 10:
                area = getMovableArea(getRname(i), stat[i][0], stat[i][1], 1, state)
                if len(area) != 0: ilist.extend(area)
    ilist2 = set (tuple ([str(x)+str(y) for x, y in ilist])) # 内包表記に近づける
    return [[int(x[0]), int(x[1])] for x in ilist2] # 各要素を分解し int 化したリスト

# これも mshogi を aids,dids 対応以外そのまま持ってくる。
# state:0(敵)、-1(何もいない)、1(Default)、2:味方を敵扱い
def getMovableAreaAllButI(i0, state): # i0 の駒がいないとして(-1なら全駒)
    global stat, aids, dids
    ilist = []
    if stat[i0][2] == 0: # i0 と同じ先後番で探索
        for i in aids:
            if i == i0: continue
            if stat[i][0] < 10:
                area = getMovableArea(getRname(i), stat[i][0], stat[i][1], 0, state)
                if len(area) != 0: ilist.extend(area)
    else:
        for i in dids:
            if i == i0: continue
            if stat[i][0] < 10:
                area = getMovableArea(getRname(i), stat[i][0], stat[i][1], 1, state)
                if len(area) != 0: ilist.extend(area)
    ilist2 = set (tuple ([str(x)+str(y) for x, y in ilist]))
    return [[int(x[0]), int(x[1])] for x in ilist2]

# この部分、速度と、筋の見逃しがトレイドオフ
def toofar(io, i, r, c):
    global stat, NAME
    [absr, absc] = [abs(stat[io][0] - r), abs(stat[io][1] - c)]
    [absrc, name] = [absr + absc, NAME[i]]
    if name == "kaku":
        if absrc > 4: return True # absrc == 3 で開き王手があったので
        else: return False
    elif name == "hisha" or name == "kyo": # 4 のケースがあったので
        if absrc > 4: return True 
        else: return False
    elif name == "fu":
        if absrc > 3: return True # 開き王手があったので
        else: return False
    elif name == "kei" and stat[i][3] == 0:
        if absr == 1 or absc == 2: return False # 開き王手は考えない
        else: return True
    else:
        if absr > 1 or absc > 1: return True # 開き王手は考えない
        else: return False

# i, r, c
def isOte(): # execEscape で使う
    global IO, stat, aids
    [r0, c0] = [stat[IO][0], stat[IO][1]]
    ote = []
    for i in aids:
        if stat[i][0] < 10:
            cango = getMovableArea(getRname(i), stat[i][0], stat[i][1], 0, 0)
            for [r, c] in cango:
                if r0 == r and c0 == c: ote.append(i)
    return ote

def execEscape(level, tejun):
    global MAXLEVEL, stat, NAME, dids, IO, kname
    # if どの手順も逃げられない: return 最長手順
    # else: return ''
    # global maxlevel, aKikiAll, IO, dstat, astat, oteIds, aMap, isTobi, aName
    if level > MAXLEVEL: return '' # turn が決まった深さに達したら探索を止める
    ote_koma = isOte() 
    if len(ote_koma) == 0: return ''
    tsumi = tsumi0 = '' # こちらは全部の手で詰まないと詰みでない
    len_tsumi = len_tsumi0 =  0
    i0 = IO
    [r0, c0, s0, n0] = stat[IO]
    sen = 1
    # ●1 王が動いてもだめか
    o_cango = getMovableArea('o', r0, c0, stat[i0][2], -1) # o が逃げられる点
    for [r1, c1] in o_cango:
        pushMap(i0, r1, c1, 1, 0)
        others_cango = getMovableAreaAll(0, 1) # もう一度相手の全攻撃エリアを取得
        if [r1, c1] in others_cango: # 相手全攻撃範囲に o がいたら、そこへは行けない
            popMap(i0)
            continue
        else: # この場合、続きを調べる必要がある
            tejun0 = tejun + str(r1) + str(c1) + '王' + ';'
            tsumi = execOte(level + 1, tejun0)
            popMap(i0)
        if tsumi == '': return ''
        else:
            len_tsumi = tsumi.count(';')
            if len_tsumi > len_tsumi0:
                tsumi0 = tsumi
                len_tsumi0 = len_tsumi
    # ●2 王手1つなら、王以外でとって判定(2つ以上の王手があれば無理なのでスルーし、次へ)
    tsumi = ''
    if (len(ote_koma) < 2):
        io = ote_koma[0] # 王手している駒の i
        [ro, co] = [stat[io][0], stat[io][1]] # 〃 座標
        for j in dids: # ote_koma を取れる味方の駒を探す
            if stat[j][0] > 9 or j == IO: continue # 条件に合わず
            can_get = getMovableArea(getRname(j), stat[j][0], stat[j][1], 1, 0)
            if [ro, co] in can_get: # 取れれば一時的にその駒で ote_koma を取る
                # 味方の全駒について
                where = "(" + str(stat[j][0]) + str(stat[j][1]) + ")"
                pushMap(io, 20, 0, 1, 0) ### トられる方を先に
                pushMap(j, ro, co, 1, stat[j][3])
                others_but = getMovableAreaAll(0, 1) # 攻撃側のキキを獲得
                if [r0, c0] in others_but: # なおも王手がかかっているのは悪手
                    popMap(j)
                    popMap(io)
                    continue
                else:
                    name = "r_" + NAME[j] if (stat[j][3] == 1) else NAME[j]
                    name = kname.get(name) + where
                    tejun0 = tejun + str(ro) + str(co) + name + ';'
                    tsumi = execOte(level + 1, tejun0)
                    popMap(j)
                    popMap(io)
            else: continue
            if tsumi == '': return ''
            else:
                len_tsumi = tsumi.count(';')
                if len_tsumi > len_tsumi0:
                    tsumi0 = tsumi
                    len_tsumi0 = len_tsumi
    # ●3 王でトル(王手が1つのとき、2つのときとも) ... 取れるコマを全て取ってチェック
    tsumi = ''
    can_get = getMovableArea('o', r0, c0, stat[i0][2], 0)
    # o が相手を取れる点
    for [r1, c1] in can_get: # 全ての o の移動について
        io = MAP[r1][c1]
        ro, co = stat[io][0], stat[io][1]
        pushMap(io, 20, 0, 1, 0)
        pushMap(i0, ro, co, 1, 0)
        others_but = getMovableAreaAll(0, 0) ##### 1 -> 0 変えると高速化
        if [r1, c1] in others_but: # 王手継続
            popMap(i0)
            popMap(io)
            continue
        else:
            tejun0 = tejun + str(r1) + str(c1) + '王' + ';'
            tsumi = execOte(level + 1, tejun0)
            popMap(i0)
            popMap(io)
        if tsumi == '': return ''
        else:
            len_tsumi = tsumi.count(';')
            if len_tsumi > len_tsumi0:
                tsumi0 = tsumi
                len_tsumi0 = len_tsumi
    # ●4 合駒ができれば合駒。途中に味方駒移動か、打ち駒なら false
    tsumi = ''
    io = ote_koma[0] # 王手している駒の i
    ro, co = stat[io][0], stat[io][1] # その位置
    if len(ote_koma) == 1: # 複数の王手なら逃れられない
        if re.match('kyo|hisha|kaku', NAME[io]): # 対応可能 kyo, hisha, kaku
            if abs(ro - r0) > 1 or abs(co - c0) > 1:
                ai = getdMochi(); # 合駒可能
                aigoma = getAigomaRC(r0, c0, ro, co)
                if len(ai) > 0: # 合駒で逃れられる
                    for i in ai:
                        for [r1, c1] in aigoma:
                            if NAME[i] == "fu":
                                if isNifu(i, r1): continue
                            pushMap(i, r1, c1, 1, 0)
                            name = kname.get(NAME[i])
                            tejun0 = tejun + str(r1) + str(c1) + name + '打;'
                            tsumi = execOte(level + 1, tejun0)
                            popMap(i)
                            if tsumi == '' or re.search("歩打;$", tsumi): return ''
                            else:
                                len_tsumi = tsumi.count(';')
                                if len_tsumi > len_tsumi0:
                                    tsumi0 = tsumi
                                    len_tsumi0 = len_tsumi
                    for j in dids: # 今度は自駒で
                        if NAME[j] != "o" and stat[j][0] < 10: # 盤面上
                            cango = getMovableArea(getRname(j), stat[j][0], \
                                                   stat[j][1], stat[j][2], -1)
                            #王以外が動ける位置
                            for [r1, c1] in aigoma:
                                if ([r1, c1] in cango): # 他の駒を動かし飛車や角の道をふさいだ
                                    where = "(" + str(stat[j][0]) + str(stat[j][1]) + ")"
                                    pushMap(j, r1, c1, 1, 0)
                                    name = kname.get(NAME[j]) + where
                                    tejun0 = tejun + str(r1) + str(c1) + name + ';'
                                    tsumi = execOte(level + 1, tejun0)
                                    popMap(j)
                                    if tsumi == '': return ''
                                    else:
                                        len_tsumi = tsumi.count(';')
                                        if len_tsumi > len_tsumi0:
                                            tsumi0 = tsumi
                                            len_tsumi0 = len_tsumi
    if tsumi0 == '': # 逃げる手が全く指せなかった場合
        if len_tsumi0 == 0:
            return tejun
    return tsumi0

# 戻り値は 動く i、場所 r1,c1, 成, 相手を取るか0, 1, 成不成 の配列
def getNext(): # 攻撃駒の次に打てる手のリストを返す
    global aids, stat, IO
    ote = []
    akiArea = getAkiArea() # 持ち駒の場合 getMochi と getAkiArea から行う
    mochi = getaMochi()
    for i in mochi:
        for [r1, c1] in akiArea:
            if isNifu(i, c1): continue
            if toofar(IO, i, r1, c1): continue # 無駄な手を減らす
            ote.append([i, r1, c1, 0, 0, ""])
    for i in aids: # 盤上の駒の場合
        if stat[i][0] < 10: # ほしいのは攻撃側の盤上の駒
            for [r1, c1] in \
                getMovableArea(getRname(i), stat[i][0], stat[i][1], stat[i][2], -1):
#                if toofar(IO, i, r1, c1): continue # 無駄な手を減らす
                if isNari(i, c1):
                    ote.append([i, r1, c1, 1, 0, ""])
                    ote.append([i, r1, c1, 0, 0, "不成"])
                else:
                    ote.append([i, r1, c1, stat[i][3], 0, ""])
            for [r1, c1] in \
                getMovableArea(getRname(i), stat[i][0], stat[i][1], stat[i][2], 0):
                # 相手を取る場合
#                if toofar(IO, i, r1, c1): continue
                if isNari(i, c1):
                    ote.append([i, r1, c1, 1, 1, ""])
                    ote.append([i, r1, c1, 0, 1, "不成"])
                else:
                    ote.append([i, r1, c1, stat[i][3], 1, ""])
    return ote

def execOte(level, tejun):
    # if ある1つの詰みを見つる: return その手順
    # else: return ''
    # 最後に execOte はできるだけ短い手順をよしとする、を入れる
    global stat, NAME, MAXLEVEL, MAP, kname
    if level > MAXLEVEL: return '' # turn が決まった深さに達したら探索を止める
    res_array = getNext()
    tsumi_insub = '' # 詰まないとしておく。どれかの手が詰めば詰み
    for [i, r1, c1, nari, catch, narazu] in res_array:
        [r0,c0] = [stat[i][0],stat[i][1]]
        if catch > 0:
            ic = MAP[r1][c1]
            if ic < 0:
                print(ic)
            pushMap(ic, 20, 0, 0, 0)
        pushMap(i, r1, c1, 0, nari)

        where = "(" + str(r0) + str(c0) + ")" if (r0 < 10) else "打"
        if (narazu == ""):
            name = "r_" + NAME[i] if (stat[i][3] == 1) else NAME[i]
            name2 = kname.get(name)
        else:
            name = NAME[i]
            name2 = kname.get(name) + narazu;
        tejun0 = tejun + str(r1) + str(c1) + name2 + where + ";";

        tsumi_insub = execEscape(level + 1, tejun0)
        popMap(i)
        if catch > 0:
            popMap(ic)
        if tsumi_insub != "" and (not re.search("歩打;$", tsumi_insub)):
            break # 詰みが発見されたら終り
    return tsumi_insub # できるだけ短い詰み

def printMAP(): # DEBUG 用の MAP 出力
    global MAP, mochi0
    ban = ""
    mochi = ""
    for i in range(1, 10):
        for j in range(1, 10):
            ban += ' %2d' % MAP[10 - j][i]
        ban += "\n"
    amochi = {'hisha':0, 'kaku':0, 'kin':0, 'gin':0, 'kei':0, 'kyo':0, 'fu':0}
    for i in aids:
        if stat[i][0] > 9:
            amochi[NAME[i]] += 1
    print(ban.rstrip())
    print(amochi)

query = ''
fname = 'Ques/0001.txt'
fname = '1.txt'
f = open(fname, 'r')
query = f.read()
f.close()

MAXLEVEL = 9 # +2手詰め
TEJUN = ""
MAP = copy.deepcopy(MAP0)
aMAP = copy.deepcopy(MAP0)
dMAP = copy.deepcopy(MAP0)
komaSet(query)

# printMAP() # デバッグのために棋譜と先手の持駒を出力する

ML = 15 # 一応 15手詰めまでを探索する

ote = isOte() # 王手以後は execEscape に任せる
if len(ote) > 0: # 0 は捨てる
    print("既に詰んでいます")
else:
    for MAXLEVEL in range(1, ML + 1, +2):
        TEJUN = execOte(0, "")
        if TEJUN != "":
            TEJUN = re.sub(';$', '', TEJUN)
            TEJUN = re.sub(';', '>', TEJUN)
            num = TEJUN.count('>')
            report = str(num + 1) + "手詰:" + TEJUN
            print(report)
            exit()
    print("手順が見つかりません")
